library jaguar_generator.swagger.builder;

import 'dart:async';

import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:source_gen/src/utils.dart';
import 'package:source_gen/src/annotation.dart';
import 'package:swagger/swagger.dart';
import 'package:yamlicious/yamlicious.dart';

import 'package:source_gen_help/source_gen_help.dart';

import 'package:jaguar/src/annotations/import.dart' as ant;

import 'package:jaguar_generator/parser/parser.dart';
import 'package:jaguar_generator/validator/validator.dart';
import 'package:jaguar_generator/swagger/to_swagger/to_swagger.dart';

class SwaggerBuilder extends Builder {
  SwaggerBuilder();

  @override
  Future build(BuildStep buildStep) async {
    AssetId id = buildStep.input.id;
    Resolver resolver = await buildStep.resolve(id, resolveAllConstants: false);
    try {
      LibraryElement lib = resolver.getLibrary(id);
      if (lib == null) return;
      await _generateForLibrary(lib, buildStep);
    } finally {
      resolver.release();
    }
  }

  @override
  List<AssetId> declareOutputs(AssetId input) {
    if (input.extension != '.dart') return const [];
    return <AssetId>[_generatedFile(input)];
  }

  AssetId _generatedFile(AssetId input, [String apiName]) {
    final String newExt = (apiName != null ? '.' + apiName : '') + '.yaml';
    return input.changeExtension(newExt);
  }

  Future _generateForLibrary(
      LibraryElement library, BuildStep buildStep) async {
    buildStep.logger.fine('Running for ${buildStep.input.id}');
    _generate(library, buildStep);
  }

  void _generate(LibraryElement unit, BuildStep buildStep) {
    for (Element element in getElementsFromLibraryElement(unit)) {
      buildStep.logger.warning('Running for $element');

      List<ElementAnnotation> matchingAnnotations =
          element.metadata.where((md) => matchAnnotation(ant.Api, md)).toList();

      if (matchingAnnotations.isEmpty) {
        continue;
      } else if (matchingAnnotations.length > 1) {
        throw 'cannot have more than one matching annotation';
      }

      ant.Api api =
          instantiateAnnotation(matchingAnnotations.single) as ant.Api;

      assert(api != null);

      if (element is! ClassElement) {
        throw new Exception("Api annotation can only be defined on a class.");
      }

      ClassElementWrap clazz = new ClassElementWrap(element);

      print(
          "Generating for api class ${clazz.name} in ${buildStep.input.id} ...");

      // Parse source code
      ParsedUpper parsed = new ParsedUpper(clazz, api.url)..parse();

      //Validate
      new ValidatorUpper(parsed)..validate();

      APIDocument doc = new ToSwagger(parsed).document;

      //TODo fix api name
      //TODO AssetId outputId = _generatedFile(buildStep.input.id, clazz.name);
      AssetId outputId = _generatedFile(buildStep.input.id);
      Asset output = new Asset(outputId, toYamlString(doc.asMap()));
      buildStep.writeAsString(output);

      buildStep.logger.finest(() => 'Generated for $element');
    }
  }

  List<AssetId> _findOutputNames(LibraryElement unit, BuildStep buildStep) {
    final List<AssetId> ret = [];
    for (Element element in getElementsFromLibraryElement(unit)) {
      buildStep.logger.warning('Running for $element');

      List<ElementAnnotation> matchingAnnotations =
          element.metadata.where((md) => matchAnnotation(ant.Api, md)).toList();

      if (matchingAnnotations.isEmpty) {
        continue;
      } else if (matchingAnnotations.length > 1) {
        throw 'cannot have more than one matching annotation';
      }

      ant.Api api =
          instantiateAnnotation(matchingAnnotations.single) as ant.Api;

      assert(api != null);

      if (element is! ClassElement) {
        throw new Exception("Api annotation can only be defined on a class.");
      }

      ClassElementWrap clazz = new ClassElementWrap(element);

      ret.add(_generatedFile(buildStep.input.id, clazz.name));
    }

    return ret;
  }
}
